"""Some custom helper types to make type hints and type checking easier."""

from pathlib import Path
from typing import Hashable, Sequence, TypeVar, Union

_Hashable = Union[Hashable, str]
path_t = TypeVar("path_t", str, Path)  # pylint:disable=invalid-name
str_t = TypeVar("str_t", str, Sequence[str])  # pylint:disable=invalid-name



from typing import Sequence
from zipfile import ZipFile, ZipInfo

from carwatch.utils._types import path_t


def filter_folder_for_participant_logs(folder_path: path_t) -> Sequence[path_t]:
    """Filter a folder for participant log files.

    The log files for one participant can either be a single zip file or a folder containing the log files.

    Parameters
    ----------
    folder_path : path or str
        path to the folder containing the log files

    Returns
    -------
    list of path or str
        list of participant logs in the folder

    """
    file_list = [
        f
        for f in sorted(folder_path.glob("*"))
        if ((f.is_file() and f.suffix == ".zip") or f.is_dir())
        and not f.name.startswith(".")  # ignore hidden files
        and not f.name.startswith("__")  # ignore hidden files
    ]
    return file_list


def filter_zip_file_for_logs(zip_ref: ZipFile) -> Sequence[ZipInfo]:
    """Filter the log files from a zip file.

    Parameters
    ----------
    zip_ref : :class:`~zipfile.ZipFile`
        zip file to filter

    Returns
    -------
    list of :class:`~zipfile.ZipInfo`
        list of log files in the zip file

    """
    file_list = [
        f
        for f in zip_ref.filelist
        if f.filename.endswith(".csv")
        and not f.filename.startswith(".")  # ignore hidden files
        and not f.filename.startswith("__")  # ignore hidden files
    ]
    return file_list



"""Module for processing CARWatch app logs."""

from carwatch.logs._participant_logs import ParticipantLogs
from carwatch.logs._study_logs import StudyLogs

__all__ = ["ParticipantLogs", "StudyLogs"]



"""A set of custom exceptions."""

__all__ = ["FileExtensionError", "LogDataParseError", "ValidationError"]


class ValidationError(Exception):
    """An error indicating that data-object does not comply with the guidelines."""


class FileExtensionError(Exception):
    """An error indicating that the file name has the wrong file extension."""


class LogDataParseError(Exception):
    """An error indicating issues while parsing log data."""



"""Module for generating QR codes to configure a study for the CARWatch app."""
import sys
from pathlib import Path
from typing import Optional, Sequence, Union

import qrcode
from qrcode.image.pil import PilImage

from carwatch.utils.study import Study
from carwatch.utils.utils import assert_is_dir, sanitize_str_for_qr


class QrCodeGenerator:
    """Class that is used to generate a QR-Code including all relevant study information for the CARWatch app.

    The QR-Code can be scanned by the CARWatch app to configure the app for a specific study.

    """

    def __init__(
        self,
        study: Study,
        saliva_distances: Union[int, Sequence[int]],
        contact_email: str,
        check_duplicates: bool = False,
        enable_manual_scan: bool = False,
    ):
        """Generate QR-Code encoding all relevant study information.

        Parameters
        ----------
        study: :obj:`~carwatch.utils.Study`
           Study object for which QR-Code will be created
        saliva_distances: int or list
            Time distances between morning samples in minutes ordered chronologically.
            For s morning saliva samples, s-1 saliva distances need to be specified.
            If the distances are constant between all morning samples,
        contact_email: str
            E-mail address that App log data will be shared with
        check_duplicates: bool
            If set to ``True``, the CAR Watch app will check and alert if a barcode has been checked twice
        enable_manual_scan: bool
            If set to ``True``, the CAR Watch app will allow manual barcode scanning apart from timed alarms

        """
        self.study = study
        self.saliva_distances = self._sanitize_distances(saliva_distances)
        self.contact = contact_email
        self.check_duplicates = check_duplicates
        self.enable_manual_scan = enable_manual_scan
        self.output_dir = None
        self.output_name = f"qr_code_{self.study.study_name}"

    def generate(self, output_dir: str = ".", output_name: Optional[str] = None):  # pragma: no cover
        """Generate a `*.png` file with QR code according to the properties of the created ``QrCodeGenerator``.

        Parameters
        ----------
        output_dir: str
            Path to the directory where the generated QR-Code will be stored
        output_name: str, optional
            Filename of the generated QR-Code.

        """
        output_dir = Path(output_dir)
        output_name = Path(output_name)
        try:
            assert_is_dir(output_dir)
        except ValueError as e:
            print(e)
            sys.exit(1)
        self.output_dir = output_dir
        if output_name:
            if output_name.suffix == ".png":
                # store without file ending
                output_name = output_name.stem
            self.output_name = output_name
        qr_img = self._generate_qr_code()
        self._save_qr_img(qr_img)

    def _generate_qr_code(self) -> PilImage:
        """Generate a QR-Code with the study information.

        This method translates the study data into a QR-Code. The information is encoded in the following format:

        ``CARWATCH;N:<study_name>;D:<num_days>;S:<subject_list>;T:<saliva_distances>;
        E:<has_evening_sample>;M:<contact_email>;F:<check_duplicates>;``

        Returns
        -------
        :obj:`qrcode.image.pil.PilImage`
            QR-Code as PIL image

        """
        # sanitize inputs to prevent decoding issues
        forbidden = [";", ":", ","]
        study_name = sanitize_str_for_qr(self.study.study_name, forbidden)
        subject_names = []
        for name in self.study.subject_names:
            sanitized_name = sanitize_str_for_qr(name, forbidden)
            subject_names.append(sanitized_name)

        app_id = "CARWATCH"
        name_string = ",".join(str(dist) for dist in subject_names)
        start_sample = 0 if self.study.start_sample_from_zero else 1
        start_sample = f"{self.study.sample_prefix}{start_sample}"
        distance_string = ",".join(str(dist) for dist in self.saliva_distances)

        # create encoding
        data = (
            f"{app_id};"
            f"N:{study_name};"
            f"D:{self.study.num_days};"
            f"S:{name_string};"
            f"SS:{start_sample};"
            f"T:{distance_string};"
            f"E:{int(self.study.has_evening_sample)};"
            f"M:{self.contact};"
            f"FD:{int(self.check_duplicates)};"
            f"FM:{int(self.enable_manual_scan)}"
        )
        # create qr code
        img = qrcode.make(data)
        return img

    def _save_qr_img(self, qr_img):
        img_location = self.output_dir.joinpath(f"{self.output_name}.png")
        qr_img.save(img_location)

    def _sanitize_distances(self, saliva_distances: Union[int, Sequence[int]]) -> Sequence[int]:
        if isinstance(saliva_distances, int):
            return [saliva_distances] * (self.study.num_samples - 1)
        if isinstance(saliva_distances, list):
            # all elements are ints
            if all(isinstance(dist, int) for dist in saliva_distances):
                # length is correct
                if len(saliva_distances) == self.study.num_samples - 1:
                    return saliva_distances
                raise ValueError(
                    f"Incorrect number of saliva distances provided! "
                    f"Needs to be {self.study.num_samples - 1}, as "
                    f"{self.study.num_samples} samples will be taken throughout the day."
                )
            raise ValueError("Invalid data detected in saliva distances! All values need to be integers!")
        raise ValueError("Saliva distances data type is invalid! Needs to be int or list of ints.")



"""Module containing example log data that can be processed."""

from pathlib import Path
from typing import Literal
from urllib.request import urlretrieve

from carwatch.logs import ParticipantLogs
from carwatch.utils._types import path_t

_EXAMPLE_DATA_PATH_LOCAL = Path(__file__).parent.parent.joinpath("example_data")
_EXAMPLE_DATA_PATH_HOME = Path.home().joinpath(".carwatch_data")
_REMOTE_DATA_PATH = "https://raw.githubusercontent.com/mad-lab-fau/carwatch/main/example_data/"


def _is_installed_manually() -> bool:
    """Check whether carwatch was installed manually and example data exists in the local path.

    Returns
    -------
    bool
        ``True`` if carwatch was installed manually, ``False`` otherwise

    """
    return (_EXAMPLE_DATA_PATH_LOCAL / "__init__.py").is_file()


def _get_data(file_name: str) -> path_t:  # pragma: no cover
    if _is_installed_manually():
        return _EXAMPLE_DATA_PATH_LOCAL.joinpath(file_name)
    path = _EXAMPLE_DATA_PATH_HOME.joinpath(file_name)
    if path.exists():
        return path
    path.parent.mkdir(parents=True, exist_ok=True)
    return _fetch_from_remote(file_name, path)


def _fetch_from_remote(file_name: str, file_path: path_t) -> path_t:  # pragma: no cover
    """Download remote dataset (helper function).

    Parameters
    ----------
    file_name : str
        file name
    file_path : str
        path to file

    Returns
    -------
    :class:`~pathlib.Path`
        path to downloaded file

    """
    url = _REMOTE_DATA_PATH + file_name
    print(f"Downloading file {file_name} from remote URL: {url}.")
    urlretrieve(url, filename=file_path)
    return file_path


def get_carwatch_log_example(participant_id: Literal["AB12C", "DE34F", "GH56I"]) -> ParticipantLogs:  # pragma: no cover
    """Get example log file.

    Parameters
    ----------
    participant_id : int
        participant ID. Must be one of ``AB12C``, ``DE34F``, or ``GH56I``.

    Returns
    -------
    :class:`~carwatch.logs.ParticipantLogs`
        log data for the selected participant

    """
    file_name = f"zip_files/logs_{participant_id}.zip"
    return ParticipantLogs.from_zip_file(_get_data(file_name))



"""Module for generating printable saliva label files."""
import importlib.resources as pkg_resources
import sys
from itertools import product
from pathlib import Path
from typing import Union

import barcode
import cairosvg
import numpy as np

from carwatch.labels.print_layout import AveryZweckformJ4791Layout, PrintLayout
from carwatch.utils import assert_is_dir, sanitize_str_for_tex, tex_to_pdf, write_to_file
from carwatch.utils.study import Study


class LabelGenerator:
    """Class that is used to generate printable `*.pdf` files with labels for all saliva samples within a study.

    Tested for labels of type `_Avery Zweckform J4791_` (48 labels per A4-sheet) and only intended to be used to print
    labels in A4 format. Note that for smaller labels, text overflows might occur.

    """

    EAN8 = barcode.get_barcode_class("ean8")
    MAX_NAME_LEN = 12

    def __init__(self, study: Study, has_barcode: bool = False, add_name: bool = False):
        """Generate printable PDF files with labels for all biomarker samples taken within a study.

        If the parameter ``add_name`` is ``True`` then the text next to the barcode will have the following layout:
        ```
        <study_name>_<subject_id>
        D<day>_<study.sample_prefix><sample_id or E for evening sample>
        ```
        For example:
        ```
        ExampleStudy_01
        D1_S4
        ```

        If ``add_name`` is set to ``False`` then only the second line is printed:
        ```
        D<day>_<study.sample_prefix><sample_id or E for evening sample>
        ```
        For example:
        ```
        D0_SE
        ```

        Parameters
        ----------
        study: :obj:`~carwatch.labels.Study`
            Study object for which labels will be created
        has_barcode: bool, optional
            Whether a barcode of type `EAN-8` will be printed on each label
        add_name: bool, optional
            Whether the name of the study will be printed on each label

        Examples
        --------
        Create a Study and print labels for a custom layout
        >>> from carwatch.labels import LabelGenerator
        >>> from carwatch.labels.print_layout import CustomLayout
        >>> from carwatch.utils import Study
        >>> Study(
        >>>     study_name="ExampleStudy", num_days=3, num_subjects=4, num_samples=5, has_evening_sample=True
        >>> )
        >>> generator = LabelGenerator(study=study, add_name=True, has_barcode=True)
        >>> layout = CustomLayout(
        >>>     num_cols=4, num_rows=10, left_margin=10, right_margin=10,
        >>>     top_margin=20, bottom_margin=20, inter_col=2, inter_row=2
        >>> )
        >>> generator.generate(output_dir="./output_labels", layout=layout)

        """
        self.study = study
        self.has_barcode = has_barcode
        self.add_name = add_name
        self.sample_prefix = sanitize_str_for_tex(self.study.sample_prefix)
        self.output_dir = None
        self.output_name = f"barcodes_{self.study.study_name}"
        self.layout = None
        self.barcode_paths = None
        self.barcodes = None
        self.barcode_ids = None

    def generate(
        self,
        output_dir: str = ".",
        output_name: Union[str, None] = None,
        layout: PrintLayout = AveryZweckformJ4791Layout(),
        debug=False,
    ):
        """Generate a `*.pdf` file with labels according to the properties of the created ``LabelGenerator``.

        Parameters
        ----------
        output_dir: str
            Path to the directory where the generated label-PDF will be stored
        output_name: str, optional
            Filename of the generated label-PDF
        layout: :obj:`~carwatch.labels.PrintLayout`, optional
            Type of label paper used for printing of the labels
        debug: bool, optional
            If set to true, auxiliary files generated by latex, e.g., `*.log`-file,
            will be preserved for debugging purpose;
            furthermore, the borders of each individual label will be drawn in the generated PDF

        """
        output_dir = Path(output_dir)
        try:
            assert_is_dir(output_dir)
        except ValueError as e:
            print(e)
            sys.exit(1)
        self.output_dir = output_dir
        if output_name:
            if output_name.endswith(".pdf"):
                # store without file ending
                output_name = output_name.rsplit(".", 1)[0]
            self.output_name = output_name
        self.layout = layout
        self._create_output_dir()
        self.barcode_ids, self.barcodes = self._generate_barcodes()
        self.barcode_paths = self._export_barcodes_to_pdf()
        self._generate_tex_file(debug=debug)
        try:
            tex_to_pdf(self.output_dir, self.output_name + ".tex")
        except RuntimeError as e:
            print(e)
        self._clear_intermediate_results(delete_img_dir=True, debug=debug)

    def _generate_barcodes(self):
        """Generate barcode ids and corresponding EAN8-codes.

        Barcode IDs and corresponding EAN8-codes are generated using the following scheme:

            `pppddss`

        where p is the participant id, d is the study day, and s is the number of the sample of that day.

        Returns
        -------
        barcode_ids: :class:`~np.array`
            1-dim array containing all 8-digit ids of the generated barcode objects
        barcodes: :class:`~np.array`
            1-dim array containing :class:`~barcode.ean.EuropeanArticleNumber8` objects that represent the EAN8-barcodes
            corresponding to ``barcode_ids``

        """
        barcode_ids = [
            f"{subj:03d}{day:02d}{saliv:02d}"
            for subj, day, saliv in list(
                product(self.study.subject_indices, self.study.day_indices, self.study.sample_indices)
            )
        ]
        # sort the ids
        barcode_ids = np.array(sorted(barcode_ids))
        # generate the codes
        barcodes = np.vectorize(self.EAN8)(barcode_ids)  # [EAN8(barcode) for barcode in barcode_ids])
        return barcode_ids, barcodes

    def _export_barcodes_to_pdf(self):
        """Create `*.svg` files from barcodes and convert them to `.pdf` files.

        The barcodes are first created as svg files from :class:`~barcode.ean.EuropeanArticleNumber8` objects
        and then converted to `*.pdf` files.

        Returns
        -------
        barcode_paths: :class:`~np.array`
            List of paths to pdf files displaying the generated barcodes

        """
        img_path = self.output_dir.joinpath("img")
        options = {"quiet_zone": 1, "font_size": 8, "text_distance": 3}  # configure label font
        barcode_paths = np.array(
            [
                barcode_object.save(img_path.joinpath(f"barcode_{barcode_id}"), options=options)
                for barcode_id, barcode_object in zip(self.barcode_ids, self.barcodes)
            ]
        )
        for path in barcode_paths:
            with Path(path).open(mode="rb") as file:
                cairosvg.svg2pdf(file_obj=file, write_to=path.replace("svg", "pdf"))
        return barcode_paths

    def _create_output_dir(self):
        """Create output directory for intermediate files or clears it (if already exists).

        The directory name will be ``{output_dir}/img``.

        """
        img_path = self.output_dir.joinpath("img")
        img_path.mkdir(exist_ok=True)
        self._clear_intermediate_results()

    def _clear_intermediate_results(self, delete_img_dir=False, debug=False):
        """Remove intermediate `*.svg` and `*.pdf` files with individual labels.

        All intermediate files are removed from ``{output_dir}/img``.
        If ``delete_img_dir`` is set to ``True``, the ``img`` folder itself will additionally be removed.

        """
        img_path = self.output_dir.joinpath("img")
        for f in img_path.glob("*"):
            if f.suffix in (".svg", ".pdf"):
                f.unlink()
        if delete_img_dir:
            img_path.rmdir()
        if not debug:
            for f in self.output_dir.glob("*"):
                if f.suffix in (".log", ".aux", ".tex"):
                    f.unlink()

    def _generate_tex_file(self, debug: bool = False):
        """Generate LaTeX code for all individual labels.

        The generated LaTeX code is then inserted into a tex-template and the result is written to
        ``{output_dir}/{output_name}.tex``.

        """
        from carwatch import labels  # package containing tex template # pylint:disable=import-outside-toplevel

        # copy tex template to output dir
        # use pkg_resources to find the template indepentently from cwd
        template = pkg_resources.read_text(labels, "barcode_template.tex")
        write_to_file(self.output_dir.joinpath(self.output_name + ".tex"), template)

        # write generated properties to output dir
        property_str = self._generate_tex_label_properties(debug)
        write_to_file(self.output_dir.joinpath("properties.tex"), property_str)

        # write generated body to output dir
        body_str = self._generate_tex_body()
        write_to_file(self.output_dir.joinpath("body.tex"), body_str)

    def _generate_label_tex(self, barcode_id: int):
        """Generate LaTeX code for a certain label.

        The label is generated based on the specified properties of :obj:`~carwatch.labels.LabelGenerator`.

        Parameters
        ----------
        barcode_id: int
            8-digit number of the respective barcode with the format `cpppddss`
            with c: check digit, p: participant id, d: study day, s: biomarker sample

        """
        study_name = sanitize_str_for_tex(self.study.study_name)
        day = int(barcode_id // 100) % 100
        sample = barcode_id % 100
        sample_name = barcode_id % 100 if self.study.start_sample_from_zero else (barcode_id + 1) % 100
        subject_index = int(barcode_id // 1e4)
        subject_name = self.study.subject_names[subject_index - 1]
        subject_name = sanitize_str_for_tex(subject_name)
        # label is realized with latex table
        label_head = "\\genericlabel\n\\begin{tabular}"
        label_foot = "\\end{tabular}\n\n"
        table_content = ""
        if self.has_barcode:
            # create table with 2 columns
            table_properties = "{m{0.45\\linewidth} m{0.4\\linewidth}}\n"
            # add barcode to first column
            table_content += (
                f"\\includegraphics[height={self.layout.get_label_height() - 4}mm,"
                f"width=\\linewidth,keepaspectratio]{{img/barcode_{subject_index:03d}{day:02d}{sample:02d}.pdf}} &"
            )

            font_size = r"\tiny"  # decrease font size to make it fit next to barcode
        else:
            # create table with only one column
            table_properties = r"{c}"
            font_size = ""  # use default font size
            # center text in label
            label_head = "\\genericlabel\n\\begin{center}\n\\begin{tabular}"
            label_foot = "\\end{tabular}\n\\end{center}\n\n"
        if self.study.has_evening_sample and all([sample == self.study.num_samples]):
            # if last sample of the day is evening sample, it is marked with "E"
            sample_name = "E"
        if self.add_name:
            delimiter = r"\_"
            if len(study_name) + len(subject_name) > LabelGenerator.MAX_NAME_LEN:
                # insert linebreak between study name and subject id to prevent overflow
                delimiter = r"\newline "
            # add study name, subject id, day, and sample to second column
            if self.has_barcode:
                # insert infos as one row in the second column
                table_content += (
                    f"{font_size}{{{study_name}{delimiter}{subject_name}"
                    f"\\newline D{day}\\_{self.sample_prefix}{sample_name}}}\n"
                )
            elif len(study_name) + len(subject_name) > LabelGenerator.MAX_NAME_LEN:
                # insert infos centered in two rows
                table_content += (
                    f"{font_size}{{{study_name}}}\\\\{{{subject_name}\\_D{day}\\_{self.sample_prefix}{sample_name}}}\n"
                )
            else:
                table_content += (
                    f"{font_size}{{{study_name}\\_{subject_name}}}\\\\{{D{day}\\_{self.sample_prefix}{sample_name}}}\n"
                )
        else:
            # add day and sample to second column
            table_content += f"\\centering{font_size}{{D{day}\\_{self.sample_prefix}{sample_name}}}\n"
        label_tex = label_head + table_properties + table_content + label_foot
        return label_tex

    def _generate_tex_label_properties(self, debug: bool) -> str:
        """Define arrangement of labels based on specified :obj:`~carwatch.labels.PrintLayout`."""
        tex_properties = ""
        tex_properties += f"\\LabelCols={self.layout.num_cols}\n"  # number of columns of labels per page
        tex_properties += f"\\LabelRows={self.layout.num_rows}\n"  # number of rows of labels per page
        tex_properties += f"\\LeftPageMargin={self.layout.left_margin}mm\n"  # margin to the left
        tex_properties += f"\\RightPageMargin={self.layout.right_margin}mm\n"  # margin to the right
        tex_properties += f"\\TopPageMargin={self.layout.top_margin}mm\n"  # margin to the top
        tex_properties += f"\\BottomPageMargin={self.layout.bottom_margin}mm\n"  # margin to the bottom
        tex_properties += f"\\InterLabelColumn={self.layout.inter_col}mm\n"  # gap between columns of labels
        tex_properties += f"\\InterLabelRow={self.layout.inter_row}mm\n"  # gap between rows of labels
        tex_properties += "\\LeftLabelBorder=1mm\n"  # minimum gap between text and label border on the left
        tex_properties += "\\RightLabelBorder=1mm\n"  # minimum gap between text and label border on the right
        tex_properties += "\\TopLabelBorder=2mm\n"  # minimum gap between text and label border on the top
        tex_properties += "\\BottomLabelBorder=2mm\n"  # minimum gap between text and label border on the bottom
        if debug:
            tex_properties += "\\LabelGridtrue\n"  # draw label borders
        return tex_properties

    def _generate_tex_body(self):
        """Iterate over all barcodes and generate corresponding LaTeX code."""
        tex_head = "\\begin{labels}\n"
        tex_foot = r"\end{labels}"
        tex_body = ""
        for sample in self.barcode_ids:
            # add tex code for every label to body
            tex_body += self._generate_label_tex(int(sample))
        return tex_head + tex_body + tex_foot



"""CARWatch: A Python package for improving the accuracy of cortisol awakening response sampling."""

__version__ = "0.3.0"

from carwatch import example_data, labels, logs

__all__ = ["labels", "logs", "example_data"]



"""Log files from a complete study."""
from __future__ import annotations

from pathlib import Path
from typing import TYPE_CHECKING, Dict, Literal, Optional, Sequence, Tuple

import pandas as pd

from carwatch.logs import ParticipantLogs
from carwatch.logs._utils import filter_folder_for_participant_logs
from carwatch.utils._types import path_t

if TYPE_CHECKING:  # pragma: no cover
    import matplotlib.pyplot as plt


class StudyLogs:
    """Class representing log files from a complete study."""

    def __init__(self, study_logs: Dict[str, ParticipantLogs]):
        """Initialize the class.

        .. note::
            This class is not intended to be used directly. Use :meth:`~carwatch.logs.StudyLogs.from_folder`
            to create a :class:`~carwatch.logs.StudyLogs` object from a folder containing log files.

        Parameters
        ----------
        study_logs : dict of str and :class:`~carwatch.logs.ParticipantLogs`
            dictionary containing the log files for each participant

        """
        self._study_logs = study_logs

    def __getitem__(self, key):
        """Return the log file for the given key."""
        return self._study_logs[key]

    def __iter__(self):
        """Iterate over the log files."""
        return iter(self._study_logs)

    def __len__(self):
        """Return the number of log files."""
        return len(self._study_logs)

    @property
    def participants(self) -> Sequence[str]:
        """Return the participants of the study."""
        return list(self._study_logs.keys())

    @classmethod
    def from_folder(
        cls,
        folder_path: path_t,
        extract_folder: Optional[bool] = False,
        overwrite_unzipped_logs: Optional[bool] = False,
        tz: Optional[str] = "Europe/Berlin",
        error_handling: Optional[Literal["ignore", "warn", "raise"]] = "raise",
    ) -> StudyLogs:
        """Create a :class:`~carwatch.logs.StudyLogs` object from a folder containing log files.

        Parameters
        ----------
        folder_path : path or str
            path to the folder containing the log files
        extract_folder : bool, optional
            ``True`` to extract the zip files to a folder, ``False`` to keep the zip files. Default: ``False``
        overwrite_unzipped_logs : bool, optional
            ``True`` to overwrite existing unzipped log files, ``False`` to keep the existing files. Default: ``False``
        tz : str, optional
            timezone of the log files. Default: "Europe/Berlin"
        error_handling : {"ignore", "warn", "raise"}, optional
            how to handle error when parse log data. ``error_handling`` can be one of the following:

            * "ignore" to ignore warning.
            * "warn" to issue warning when no "Subject ID Set" action was found in the data (indicating that a
              participant did not correctly register itself for the study or that log data is corrupted)
            * "raise" to raise an exception when no "Subject ID Set" action was found in the data.
            Default: "ignore"

        Returns
        -------
        :class:`~carwatch.logs.StudyLogs`
            :class:`~carwatch.logs.StudyLogs` object containing the log files

        """
        folder_path = Path(folder_path)
        study_logs = {}
        file_list = filter_folder_for_participant_logs(folder_path)

        if len(file_list) == 0:
            raise FileNotFoundError(f"No log files found in folder {folder_path}.")

        for path in file_list:
            if path.is_file() and path.suffix == ".zip":
                logs = ParticipantLogs.from_zip_file(
                    path,
                    extract_folder=extract_folder,
                    overwrite_unzipped_logs=overwrite_unzipped_logs,
                    tz=tz,
                    error_handling=error_handling,
                )
            else:  # must be a folder
                logs = ParticipantLogs.from_folder(path, tz=tz, error_handling=error_handling)
            study_logs[logs.subject_id] = logs

        return cls(study_logs)

    def data_as_df(self):
        """Return the log files as a :class:`~pandas.DataFrame`.

        Returns
        -------
        :class:`~pandas.DataFrame`
            log files as a :class:`~pandas.DataFrame`

        """
        return pd.concat(
            {key: participant_logs.data_as_df() for key, participant_logs in self._study_logs.items()},
            names=["subject"],
        )

    def export_times(
        self,
        sampling_times: Optional[bool] = True,
        awakening_times: Optional[bool] = True,
        include_evening_sample: Optional[bool] = True,
        wide_format: Optional[bool] = False,
    ) -> pd.DataFrame:
        """Export the sampling and/or awakening times of all log files from the study as a :class:`~pandas.DataFrame`.

        Parameters
        ----------
        sampling_times : bool, optional
            ``True`` to include sampling times, ``False`` to exclude. Default: ``True``
        awakening_times : bool, optional
            ``True`` to include awakening times, ``False`` to exclude. Default: ``True``
        include_evening_sample : bool, optional
            ``True`` to include evening sampling times, ``False`` to only include morning sampling times.
            Default: ``True``
        wide_format : bool, optional
            ``True`` to return a wide format dataframe where one row represents one participant, ``False`` to return a
            dataframe where one row represents one day/night. Default: ``False``

        Returns
        -------
        :class:`~pandas.DataFrame`
            sampling and/or awakening times of all log files from the study as a :class:`~pandas.DataFrame`

        """
        data_dict = {
            key: participant_logs.export_times(
                sampling_times=sampling_times,
                awakening_times=awakening_times,
                include_evening_sample=include_evening_sample,
                wide_format=wide_format,
            )
            for key, participant_logs in self._study_logs.items()
        }
        if wide_format:
            return pd.concat(data_dict.values())

        return pd.concat(data_dict, names=["subject"])

    @property
    def android_versions(self) -> pd.DataFrame:
        """Return the Android versions of the different study participants as a :class:`~pandas.DataFrame`.

        Returns
        -------
        :class:`~pandas.DataFrame`
            Android versions of the different study participants as a :class:`~pandas.DataFrame`
        """
        return self._get_metadata("android_version")

    @property
    def app_versions(self):
        """Return the app versions of the different study participants as a :class:`~pandas.DataFrame`.

        Returns
        -------
        :class:`~pandas.DataFrame`
            App versions of the different study participants as a :class:`~pandas.DataFrame`
        """
        return self._get_metadata("app_version")

    @property
    def phone_models(self):
        """Return the phone models of the different study participants as a :class:`~pandas.DataFrame`.

        Returns
        -------
        :class:`~pandas.DataFrame`
            Phone models of the different study participants as a :class:`~pandas.DataFrame`
        """
        return self._get_metadata("phone_model")

    @property
    def phone_manufacturers(self):
        """Return the phone manufacturers of the different study participants as a :class:`~pandas.DataFrame`.

        Returns
        -------
        :class:`~pandas.DataFrame`
            Phone manufacturers of the different study participants as a :class:`~pandas.DataFrame`
        """
        return self._get_metadata("phone_manufacturer")

    def _get_metadata(self, metadata_type: str) -> pd.DataFrame:
        data = {key: getattr(logs, metadata_type) for key, logs in self._study_logs.items()}
        data = pd.Series(data).to_frame(name=metadata_type)
        data.index.name = "subject"
        return data

    def get_metadata_stats(self, metadata: str) -> pd.DataFrame:
        """Return the number of participants for each value of the given metadata.

        Parameters
        ----------
        metadata : str
            metadata to get the number of participants for each value of

        Returns
        -------
        :class:`~pandas.DataFrame`
            number of participants for each value of the given metadata

        """
        return pd.DataFrame(self._get_metadata(metadata).value_counts(), columns=["count"])

    def get_metadata_plot(self, metadata: str, **kwargs) -> Tuple[plt.Figure, plt.Axes]:  # pragma: no cover
        try:
            import matplotlib.pyplot as plt
            import seaborn as sns
        except ImportError as e:
            raise ImportError(
                "This function requires matplotlib and seaborn to be installed. "
                "Please install them either manually or by installing the 'carwatch' package "
                "using 'pip install carwatch[plotting]'."
            ) from e
        ax = kwargs.pop("ax", None)
        if ax is None:
            fig, ax = plt.subplots(figsize=kwargs.pop("figsize", None))
        else:
            fig = ax.get_figure()

        data = self.get_metadata_stats(metadata)

        palette = kwargs.pop("palette", sns.cubehelix_palette(len(data), start=0.5, rot=-0.75))
        ax = sns.barplot(data=data.reset_index(), x=metadata, y="count", ax=ax, palette=palette, **kwargs)

        if metadata in ["phone_model", "log_dates"]:
            ax.set_xticklabels(ax.get_xticklabels(), rotation=45, ha="right")

        return fig, ax



"""Log files from one single participant."""
import json
import warnings
import zipfile
from itertools import product
from pathlib import Path
from typing import IO, Any, Dict, Literal, Optional, Sequence, Union

import numpy as np
import pandas as pd

from carwatch.logs._utils import filter_zip_file_for_logs
from carwatch.utils._datatype_validation_helper import _assert_file_extension, _assert_is_dtype
from carwatch.utils._types import path_t, str_t
from carwatch.utils.exceptions import LogDataParseError


class ParticipantLogs:
    """Class representing log data from one single participant."""

    log_actions: Dict[str, Sequence[str]] = {
        "app_metadata": ["app_version_code", "app_version_name"],
        "phone_metadata": [
            "brand",
            "manufacturer",
            "model",
            "version_sdk_level",
            "version_security_patch",
            "version_release",
        ],
        "subject_id_set": ["subject_id", "subject_condition"],
        "alarm_set": ["alarm_id", "timestamp", "is_repeating", "is_hidden", "hidden_timestamp"],
        "timer_set": ["alarm_id", "timestamp"],
        "alarm_cancel": ["alarm_id"],
        "alarm_ring": ["alarm_id", "saliva_id"],
        "alarm_snooze": ["alarm_id", "snooze_duration", "source"],
        "alarm_stop": ["alarm_id", "source", "saliva_id"],
        "alarm_killall": [],
        "evening_salivette": ["alarm_id"],
        "barcode_scan_init": [],
        "barcode_scanned": ["alarm_id", "saliva_id", "barcode_value"],
        "invalid_barcode_scanned": ["barcode_value"],
        "duplicate_barcode_scanned": ["barcode_value", "other_barcodes"],
        "spontaneous_awakening": ["alarm_id"],
        "lights_out": [],
        "day_finished": ["day_counter"],
        "service_started": [],
        "service_stopped": [],
        "screen_off": [],
        "screen_on": [],
        "user_present": [],
        "phone_boot_init": [],
        "phone_boot_complete": [],
    }

    def __init__(self, data: pd.DataFrame, tz: str, error_handling: str = "ignore") -> None:
        """Initialize new ``ParticipantLogs`` instance.

        .. note::
            Usually you shouldn't use this init directly.
            Use the provided `from_zip_file` or `from_folder` constructors to load CARWatch log data.

        Parameters
        ----------
        data : :class:`~pandas.DataFrame`
            log data as dataframe
        tz : str
            timezone of the log data
        error_handling : {"ignore", "warn", "raise"}, optional
            how to handle error when parsing log data. ``error_handling`` can be one of the following:
            ``"ignore"``: ignore errors and continue parsing
            ``"warn"``: ignore errors and continue parsing, but print a warning
            ``"raise"``: raise an exception when an error occurs

        """
        self._data: pd.DataFrame = data
        self.tz: str = tz
        self.subject_id: str
        self.error_handling: str = error_handling
        self.log_dates: Optional[Sequence[pd.Timestamp]] = None
        self.app_metadata: Optional[Dict[str, Any]] = None
        self.phone_metadata: Optional[Dict[str, Any]] = None
        self._extract_info()

    @classmethod
    def from_zip_file(
        cls,
        path: path_t,
        extract_folder: Optional[bool] = False,
        overwrite_unzipped_logs: Optional[bool] = False,
        tz: Optional[str] = "Europe/Berlin",
        error_handling: Optional[Literal["ignore", "warn", "raise"]] = "raise",
    ) -> "ParticipantLogs":
        """Load log files from one subject.

        Parameters
        ----------
        path : :class:`~pathlib.Path` or str
            path to zip file from one participant
        extract_folder : bool, optional
            ``True`` to extract zip file to a folder with the same name as the zip file,
            ``False`` to only read content and not extract. Default: ``False``
        overwrite_unzipped_logs : bool, optional
            ``True`` to overwrite already unzipped log files, ``False`` to not overwrite.
            Only relevant if ``extract_folder`` is ``True``. Default: ``False``
        error_handling : {"ignore", "warn", "raise"}, optional
            how to handle error when parse log data. ``error_handling`` can be one of the following:

            * "ignore" to ignore warning.
            * "warn" to issue warning when no "Subject ID Set" action was found in the data (indicating that a
              participant did not correctly register itself for the study or that log data is corrupted)
            * "raise" to raise an exception when no "Subject ID Set" action was found in the data.
            Default: "ignore"
        tz : str, optional
            timezone of the log data. Default: "Europe/Berlin"

        Returns
        -------
        :class:`~carwatch.logs.ParticipantLogs`
            log data from one participant

        """
        path = Path(path)

        _assert_file_extension(path, ".zip")
        if not path.exists():
            raise FileNotFoundError(f"File {path} does not exist.")

        with zipfile.ZipFile(path, "r") as zip_ref:
            if extract_folder:
                # extract zip file to folder with same name as zip file
                export_folder = path.parent.joinpath(path.stem)
                export_folder.mkdir(exist_ok=True)
                if overwrite_unzipped_logs or len(list(export_folder.glob("*"))) == 0:
                    zip_ref.extractall(export_folder)
                else:
                    # folder not empty => inform user and load folder
                    warnings.warn(
                        f"Folder {export_folder.name} already contains log files which will be loaded. "
                        f"Set `overwrite_logs_unzip = True` to overwrite log files."
                    )
                log_data = cls._folder_to_dataframe(export_folder, tz)
            else:
                log_data = cls._zip_file_to_dataframe(zip_ref, tz)

        return cls(log_data, tz, error_handling)

    @classmethod
    def from_folder(
        cls,
        folder_path: path_t,
        tz: Optional[str] = "Europe/Berlin",
        error_handling: Optional[Literal["ignore", "warn", "raise"]] = "raise",
    ) -> "ParticipantLogs":
        """Load log files from one participant.

        Parameters
        ----------
        folder_path : :class:`~pathlib.Path` or str
            path to folder containing log files from one participant
        tz : str
            timezone of the log data
        error_handling : {"ignore", "warn", "raise"}, optional
            how to handle error when parsing log data. ``error_handling`` can be one of the following:

            * "ignore" to ignore warning.
            * "warn" to issue warning when no "Subject ID Set" action was found in the data (indicating that a
                participant did not correctly register itself for the study or that log data is corrupted)
            * "raise" to raise an exception when no "Subject ID Set" action was found in the data.
            Default: "ignore"


        Returns
        -------
        :class:`~carwatch.logs.ParticipantLogs`
            log data from one participant

        """
        folder_path = Path(folder_path)
        log_data = cls._folder_to_dataframe(folder_path, tz, error_handling=error_handling)
        return cls(log_data, tz, error_handling=error_handling)

    @classmethod
    def _folder_to_dataframe(
        cls, folder_path: Path, tz: str, error_handling: Optional[Literal["ignore", "warn", "raise"]] = "raise"
    ) -> pd.DataFrame:
        """Load log data from folder of one participant and return it as dataframe.

        Parameters
        ----------
        folder_path : :class:`~pathlib.Path` or str
            path to folder containing log files from participant
        tz : str
            timezone of the log data
        error_handling : {"ignore", "warn", "raise"}, optional
            how to handle error when parsing log data. ``error_handling`` can be one of the following:


        Returns
        -------
        :class:`~pandas.DataFrame`
            dataframe with log data for one participant

        """
        folder_path = Path(folder_path)
        if not folder_path.exists():
            raise FileNotFoundError(f"Folder {folder_path} does not exist.")
        file_list = sorted(folder_path.glob("*.csv"))
        if len(file_list) == 0:
            raise FileNotFoundError(f"No log files found in folder {folder_path}.")
        return pd.concat([cls._load_log_file_csv(file, tz, error_handling=error_handling) for file in file_list])

    @classmethod
    def _zip_file_to_dataframe(cls, zip_ref: zipfile.ZipFile, tz: str) -> pd.DataFrame:
        """Load log data from zip file of one participant and return it as dataframe.

        Parameters
        ----------
        zip_ref : :class:`~zipfile.ZipFile`
            zip file containing log files from participant
        tz : str
            timezone of the log data

        Returns
        -------
        :class:`~pandas.DataFrame`
            dataframe with log data for one participant

        """
        file_list = filter_zip_file_for_logs(zip_ref)
        return pd.concat([cls._load_log_file_csv(zip_ref.open(file), tz) for file in file_list])

    @classmethod
    def _load_log_file_csv(
        cls,
        file_path: Union[Path, IO[bytes]],
        tz: str,
        error_handling: Optional[Literal["ignore", "warn", "raise"]] = "raise",
    ) -> pd.DataFrame:
        df = pd.read_csv(file_path, sep=";", header=None, names=["time", "action", "extras"])

        # assert that time is monotonic increasing
        if not df["time"].is_monotonic_increasing and error_handling in ["warn", "raise"]:
            msg = (
                "The timestamps in the log data are not monotonic increasing. "
                "It might be corrupted or manually tampered with. If you want to load the log data anyway, "
                "set `error_handling = 'ignore'`."
            )
            if error_handling == "warn":
                warnings.warn(msg)
            else:
                raise LogDataParseError(msg)

        df["time"] = pd.to_datetime(df["time"], unit="ms")
        df = df.set_index("time")
        df.index = df.index.tz_localize("UTC").tz_convert(tz)
        df = df.sort_index()
        df = df.apply(cls._parse_date, args=(tz,), axis=1)
        return df

    @classmethod
    def _parse_date(cls, row: pd.Series, tz: str) -> pd.Series:
        """Parse date in "timestamp" and "timestamp_hidden" columns of a pandas series.

        Parameters
        ----------
        row : :class:`~pandas.Series`
            one row of a dataframe
        tz : str
            timezone of the log data

        Returns
        -------
        :class:`~pandas.Series`
            series with parsed date

        """
        json_extra = json.loads(row["extras"])
        row_cpy = row.copy()
        keys = ["timestamp", "timestamp_hidden"]
        for key in keys:
            if key in json_extra:
                # convert to datetime and localize
                time = pd.to_datetime(json_extra[key], unit="ms")
                time = time.tz_localize("UTC").tz_convert(tz)
                json_extra[key] = str(time)

        # fix wrong key for saliva_id in "alarm_ring" action (old app versions)
        if "extra_saliva_id" in json_extra:
            json_extra["saliva_id"] = json_extra["extra_saliva_id"]
            del json_extra["extra_saliva_id"]
        row_cpy["extras"] = json.dumps(json_extra)
        return row_cpy

    def _extract_info(self) -> None:
        """Extract log data information."""
        # Subject Information
        subject_dict = self.get_extras_for_log_action("subject_id_set")
        try:
            self.subject_id = subject_dict["subject_id"]
        except KeyError as e:
            if self.error_handling == "warn":
                warnings.warn("Action 'Subject ID Set' not found - Log Data may be invalid!")
            elif self.error_handling == "raise":
                raise LogDataParseError("Action 'Subject ID Set' not found - Log Data may be invalid!") from e
        # App Metadata
        self.app_metadata = self.get_extras_for_log_action("app_metadata")
        # Phone Metadata
        self.phone_metadata = self.get_extras_for_log_action("phone_metadata")
        # Study Metadata
        # TODO hardcoded for now, change after study metadata is implemented
        self.study_metadata = {
            "study_name": "Test",
            "study_type": "CAR",
            "num_days": 2,
            "saliva_ids": {0: "S1", 1: "S2", 2: "S3", 3: "S4", 4: "S5", 5: "SA"},
            "time_points": [0, 15, 30, 45, 60],
            "has_evening_sample": True,
        }
        #  self.study_metadata = self.get_extras_for_log_action("study_metadata")

        # Log Info
        self.log_dates = np.array(list(self._data.index.normalize().unique()))

    #
    #     def _ipython_display_(self):
    #         self.print_info()
    #
    #     def print_info(self):
    #         """Display Markdown-formatted log data information."""
    #         try:
    #             from IPython.core.display import Markdown, display  # pylint:disable=import-outside-toplevel
    #         except ImportError as e:
    #             raise ImportError(
    #                 "Displaying LogData information failed because "
    #                 "IPython cannot be imported. Install it via 'pip install ipython'."
    #             ) from e
    #
    #         display(Markdown("Subject ID: **{}**".format(self.subject_id)))
    #         display(Markdown("Condition: **{}**".format(self.condition)))
    #         display(Markdown("App Version: **{}**".format(self.app_version)))
    #         display(Markdown("Android Version: **{}**".format(self.android_version)))
    #         display(Markdown("Phone: **{}**".format(self.model)))
    #         display(Markdown("Logging Days: **{} - {}**".format(str(self.start_date), str(self.end_date))))
    #

    @property
    def android_version(self) -> int:
        """Return the Android version.

        Returns
        -------
        int
            SDK version of Android version or 0 if information is not available

        """
        return self.phone_metadata.get("version_sdk_level", 0)

    @property
    def app_version(self) -> str:
        """Return the CARWatch App version.

        Returns
        -------
        str
            CARWatch App version

        """
        return self.app_metadata.get("version_name").split("_")[0]

    @property
    def phone_manufacturer(self) -> str:
        """Return the phone manufacturer.

        Returns
        -------
        str
            name of the phone manufacturer or "n/a" if information is not available

        """
        return self.phone_metadata.get("manufacturer", "n/a")

    @property
    def phone_model(self) -> str:
        """Return the phone model.

        Returns
        -------
        str
            name of the phone model or "n/a" if information is not available

        """
        return self.phone_metadata.get("model", "n/a")

    @property
    def start_date(self) -> pd.Timestamp:
        """Return the start date of log data, i.e., the first day with log data.

        ..note:: This is most likely the day the app was installed and configured and not necessarily the date
        when the CAR procedure was started.

        Returns
        -------
        :class:`~pandas.Timestamp`
            start date


        """
        return self.log_dates[0]

    @property
    def end_date(self) -> pd.Timestamp:
        """Return end date of log data.

        ..note:: This is not necessarily the date when the CAR procedure was finished.

        Returns
        -------
        :class:`~pandas.Timestamp`
            start date

        """
        return self.log_dates[-1]

    @property
    def num_saliva_samples(self) -> int:
        """Return number of saliva samples.

        Returns
        -------
        int
            number of saliva samples

        """
        return len(self.study_metadata.get("time_points"))

    @property
    def saliva_ids(self) -> Dict[int, str]:
        """Return saliva ids.

        Returns
        -------
        dict of int, str
            mapping of saliva indices to saliva ids

        """
        return self.study_metadata.get("saliva_ids")

    @property
    def time_points(self) -> Sequence[int]:
        """Return saliva sample time points.

        Returns
        -------
        list of int
            saliva sample time points in minutes

        """
        return self.study_metadata.get("time_points")

    @property
    def study_type(self) -> str:
        """Return study type.

        Returns
        -------
        str
            study type

        """
        return self.study_metadata.get("study_type")

    @property
    def has_evening_sample(self) -> bool:
        """Return whether the study has an evening saliva sample.

        Returns
        -------
        bool
            whether the study has an evening saliva sample

        """
        return self.study_metadata.get("has_evening_sample")

    def data_as_df(self) -> pd.DataFrame:
        """Return log data as dataframe.

        Returns
        -------
        :class:`~pandas.DataFrame`
            log data

        """
        return self._data.copy()

    def filter_logs(
        self,
        *,
        data: Optional[pd.DataFrame] = None,
        action: Optional[str_t] = None,
        date: Optional[Union[pd.Timestamp, str]] = None,
    ) -> pd.DataFrame:
        """Filter logs by action and date.

        Parameters
        ----------
        data : :class:`~pandas.DataFrame`, optional
            dataframe with log data to filter or ``None`` to use the log data of the current instance. Default: ``None``
        action : str or list of str, optional
            action(s) to filter log data for or ``None`` to filter for all actions. Default: ``None``
        date : :class:`~pandas.Timestamp` or str, optional
            date to filter log data for or ``None`` to filter for all dates. Default: ``None``

        Returns
        -------
        :class:`~pandas.DataFrame`
            dataframe with filtered log data

        """
        return self._get_logs_for_action(data=data, action=action, date=date)

    def _get_logs_for_date(self, data: pd.DataFrame, date: Union[str, pd.Timestamp]) -> pd.DataFrame:
        """Filter log data for a specific date.

        Parameters
        ----------
        data : :class:`~pandas.DataFrame`
            dataframe with log data to filter
        date : str or :class:`~pandas.Timestamp`
            date to filter log data for

        Returns
        -------
        :class:`~pandas.DataFrame`
            dataframe with log data for specific date

        """
        date = pd.Timestamp(date).tz_localize(self.tz)

        if date is pd.NaT:
            return data

        return data.loc[data.index.normalize() == date]

    def split_sampling_days(
        self, split_into_nights: Optional[bool] = True, return_dict: Optional[bool] = False
    ) -> Union[pd.DataFrame, Dict[str, pd.DataFrame]]:
        """Split continuous log data into individual sampling days.

        This function splits data into individual sampling days.
        The split is performed at 6pm because that's the time of day when the probability of
        sleeping is the lowest.

        Parameters
        ----------
        split_into_nights : bool, optional
            If ``True``, split data into *nights*, assuming that samples taken in the evening correspond to the
            next day. This means that data are split at 6pm. This is the typical way if cortisol awakening response
            (CAR) data with preceding evening saliva samples are assessed.
            If ``False``, split data into *days*, assuming that all samples taken at one day belong to the same day.
            This is the typical way if cortisol daily profiles with cortisol samples over the whole day are assessed.
            Default: ``True``
        return_dict : bool, optional
            If ``True``, return a dictionary with sampling date (keys) and log dataframes (values). If ``False``,
            a concatenated dataframe is returned. Default: ``False``


        Returns
        -------
        dict
            dictionary with sampling date (keys) and log dataframes (values).

        """
        data = self.data_as_df()
        _assert_is_dtype(data.index, pd.DatetimeIndex)
        # split data per day
        date_diff = np.diff(data.index.date)
        date_diff = np.append(date_diff[0], date_diff)
        idx_date = np.where(date_diff)[0]

        dict_data = self._split_night(data, idx_date) if split_into_nights else self._split_day(data, idx_date)

        if return_dict:
            return dict_data

        return pd.concat(dict_data, names=["date"])

    @staticmethod
    def _split_night(data: pd.DataFrame, idx_date: np.ndarray) -> Dict[str, pd.DataFrame]:
        # split data per time: split at 6 pm, because that's the time of day
        # when the probability of sleeping is the lowest
        time_6pm = pd.Timestamp("18:00:00").time()

        time_diff = data.index.time <= time_6pm
        time_diff = np.append(time_diff[0], time_diff)
        idx_time = np.where(np.diff(time_diff))[0]

        # concatenate both splitting criteria and split data
        idx_split = np.unique(np.concatenate([idx_date, idx_time]))
        data_split = np.split(data, idx_split)

        # concatenate data from one night (data between 6 pm and 12 am from the previous day and
        # between 12 am and 6 pm of the next day)
        for i, df in enumerate(data_split):
            if i < (len(data_split) - 1):
                df_curr = df
                df_next = data_split[i + 1]

                date_curr = df_curr.index[0].date()
                date_next = df_next.index[0].date()
                time_curr = df_curr.index[0].time()
                time_next = df_next.index[0].time()

                # check if dates are consecutive and if first part is after 6 pm and second part is before 6 am
                if (date_next == date_curr + pd.Timedelta("1d")) and (time_curr > time_6pm > time_next):
                    data_split[i] = pd.concat([df_curr, df_next])
                    # delete the second part
                    del data_split[i + 1]

        # create dict with data from each night. dictionary keys are the dates.
        dict_data = {}
        for df in data_split:
            date = df.index[-1].normalize().date()
            # By convention, if the recording started after 6 pm the date is assumed to be corresponding
            # to the next morning
            if df.index[-1].time() > time_6pm:
                date += pd.Timedelta("1d")
            dict_data[str(date)] = df

        return dict_data

    @staticmethod
    def _split_day(data, idx_date):
        data_split = np.split(data, idx_date)
        dict_data = {}
        for df in data_split:
            date = df.index[0].normalize().date()
            dict_data[str(date)] = df

        return dict_data

    def _get_logs_for_action(
        self,
        *,
        data: Optional[pd.DataFrame] = None,
        action: Optional[str_t] = None,
        date: Optional[Union[pd.Timestamp, str]] = None,
        rows: Optional[Union[str, int, Sequence[int]]] = None,
    ) -> Union[pd.DataFrame, pd.Series]:
        """Filter log data for a specific action.

        Parameters
        ----------
        data : :class:`~pandas.DataFrame`, optional
            dataframe with log data to filter or ``None`` to use the log data of the current instance. Default: ``None``
        action : str or list of str, optional
            action(s) to filter log data for or ``None`` to filter for all actions. Default: ``None``
        date : :class:`~pandas.Timestamp` or str, optional
            filter log data to only contain data from one selected date or ``None`` to include data from all dates.
            Default: ``None``
        rows : str, int, or list of int, optional
            index label (or list of such) to slice filtered log data (e.g., only select the first action) or
            ``None`` to include all data. Default: ``None``

        Returns
        -------
        :class:`~pandas.DataFrame`
            dataframe with log data for specific action

        """
        if data is None:
            data = self.data_as_df()

        if date is not None:
            data = self._get_logs_for_date(data=data, date=date)

        if isinstance(action, str):
            action = [action]
        if action is None:
            return data
        if all(a not in self.log_actions.keys() for a in action):
            return pd.DataFrame()

        data = data.set_index("action", append=True)
        data_filter = data.reindex(action, level="action").reset_index(level="action")
        if rows:
            data_filter = data_filter.iloc[rows, :]
        return data_filter

    def get_extras_for_log_action(self, action: str) -> Dict[str, str]:
        """Extract log data extras from log data.

        Parameters
        ----------
        action : :class:`datetime.date` or str
            action to filter log data

        Returns
        -------
        dict
            dictionary with log extras for specific action

        """
        row = self._get_logs_for_action(action=action, rows=0)
        if row.empty:
            return {}

        return json.loads(row["extras"].iloc[0])

    def sampling_times(self, include_evening_sample: Optional[bool] = True) -> pd.DataFrame:
        """Extract saliva sampling times from log data.

        Parameters
        ----------
        include_evening_sample : bool, optional
            ``True`` to include evening sampling times, ``False`` to only include morning sampling times.
            Default: ``True``

        Returns
        -------
        :class:`~pandas.DataFrame`
            dataframe with sampling times

        """
        data_split = {}
        for day, data in self.split_sampling_days(return_dict=True).items():
            data_filt = self.filter_logs(data=data, action=["barcode_scanned"])
            if len(data_filt) == 0:
                continue
            # expand extras, which is a json string
            extras = data_filt["extras"].apply(json.loads)
            # convert extras to dataframe
            df = pd.DataFrame(list(extras.values))
            df = df[["saliva_id"]]

            if self.study_type == "CAR" and self.has_evening_sample:
                # assign morning or evening "saliva_type" to each saliva_id
                df = df.assign(
                    timestamp=data_filt.index, sampling_time=data_filt.index.strftime("%H:%M:%S"), saliva_type="morning"
                )
                # the last saliva_id is the evening sample
                df.loc[df["saliva_id"] == list(self.saliva_ids.keys())[-1], "saliva_type"] = "evening"

            # find the last barcode_scanned event for each saliva_id
            df = df.groupby("saliva_id", sort=False).last()
            df = df.reset_index().set_index(["saliva_type", "saliva_id"])
            df = df.rename(self.saliva_ids, level="saliva_id")
            data_split[day] = df

        data_concat = pd.concat(data_split, names=["date"])

        if not include_evening_sample:
            data_concat = data_concat.drop("evening", level="saliva_type")

        # assign a unique id to each night, starting with 1
        date_vals = data_concat.index.get_level_values("date")
        data_concat = data_concat.assign(day_id=date_vals.unique().searchsorted(date_vals) + 1)
        data_concat = data_concat.set_index("day_id", append=True)
        return data_concat

    def awakening_times(self) -> pd.DataFrame:
        """Extract awakening times from log data.

        Returns
        -------
        :class:`~pandas.DataFrame`
            dataframe with awakening times

        """
        data_split = {}
        for day, data in self.split_sampling_days(return_dict=True).items():
            data_action = []
            for log_action in ["spontaneous_awakening", "alarm_stop"]:
                data_filter = self.filter_logs(data=data, action=log_action)
                if len(data_filter) == 0:
                    continue
                # expand extras, which is a json string
                extras = data_filter["extras"].apply(json.loads)
                # convert extras to dataframe
                df = pd.DataFrame(list(extras.values))
                df.index = data_filter.index
                df = data_filter.drop(columns=["extras"]).join(df)
                df = df.assign(awakening_time=data_filter.index.strftime("%H:%M:%S"))
                df = df.set_index("action", append=True)
                data_action.append(df)

            if len(data_action) == 0:
                continue
            data_day = pd.concat(data_action)
            data_day = data_day.iloc[[0], :]
            data_day = data_day.reset_index("action")
            data_day = data_day.rename(columns={"action": "awakening_type"})
            data_day["awakening_type"] = data_day["awakening_type"].replace(
                {"spontaneous_awakening": "self-report", "alarm_stop": "alarm"}
            )
            data_day.index.name = "timestamp"

            # todo test
            if data_day.iloc[0]["awakening_type"] == "alarm" and data_day["saliva_id"].iloc[0] not in [np.nan, 0]:
                data_day = data_day.reset_index()
                data_day.iloc[0, :] = np.nan
                data_day = data_day.set_index("timestamp")

            data_day = data_day[["awakening_time", "awakening_type"]]
            data_split[day] = data_day

        data_concat = pd.concat(data_split, names=["date"])
        data_concat = data_concat.reset_index("timestamp")

        data_concat = data_concat.assign(day_id=range(1, len(data_concat) + 1))
        data_concat = data_concat.set_index("day_id", append=True)
        return data_concat

    def export_times(
        self,
        sampling_times: Optional[bool] = True,
        awakening_times: Optional[bool] = True,
        include_evening_sample: Optional[bool] = True,
        wide_format: Optional[bool] = False,
    ) -> pd.DataFrame:
        """Export sampling and/or awakening times from the participant's logs.

        Parameters
        ----------
        sampling_times : bool, optional
            ``True`` to include sampling times, ``False`` to exclude. Default: ``True``
        awakening_times : bool, optional
            ``True`` to include awakening times, ``False`` to exclude. Default: ``True``
        include_evening_sample : bool, optional
            ``True`` to include evening sampling times, ``False`` to only include morning sampling times.
            Default: ``True``
        wide_format : bool, optional
            ``True`` to return a wide format dataframe where one row represents one participant, ``False`` to return a
            dataframe where one row represents one day/night. Default: ``False``

        Returns
        -------
        :class:`~pandas.DataFrame`
            dataframe with sampling and/or awakening times

        """
        list_data = []
        if awakening_times:
            data = self.awakening_times()
            data = data[["awakening_time", "awakening_type"]]
            list_data.append(data)
        if sampling_times:
            data = self.sampling_times(include_evening_sample=include_evening_sample)
            data = data[["sampling_time"]]
            data = data.droplevel("saliva_type")
            data = data.unstack(["saliva_id"])
            data.columns = ["_".join(map(str, col)) for col in data.columns]
            list_data.append(data)

        if len(list_data) == 0:
            raise ValueError("No data to export.")

        data_concat = pd.concat(list_data, axis=1)
        if wide_format:
            data_concat = data_concat.reset_index("date").unstack("day_id")
            data_concat = pd.DataFrame(data_concat, columns=[self.subject_id]).T

            # rearrange columns so that "date_*" is first, followed by "awakening_*", then "sampling_*"
            col_names = list(data_concat.columns.get_level_values(level=0).unique())
            day_names = list(data_concat.columns.get_level_values(level="day_id").unique())
            col_names = list(
                product(
                    day_names,
                    col_names,
                )
            )
            # invert order of tuples
            col_names = [(col[1], col[0]) for col in col_names]
            data_concat = data_concat.reindex(columns=col_names)
            data_concat.columns = [f"{col[0]}_D{col[1]}" for col in data_concat.columns]
            data_concat.index.name = "subject"

        return data_concat



"""Main script for the CLI interface of the carwatch package."""
import itertools
import sys
import threading
import time
from typing import Optional, Sequence, Union

import click
from trogon import tui

from carwatch.labels import LabelGenerator
from carwatch.labels.print_layout import CustomLayout
from carwatch.qr_codes import QrCodeGenerator
from carwatch.utils import Condition, Study, validate_mail_input, validate_saliva_distances, validate_subject_path
from carwatch.utils.click_helper import NumericChoice, get_file_name

CAR_STUDY = 1
LAB_STUDY = 2
OTHER_STUDY = 3
STUDY_TYPES = [CAR_STUDY, LAB_STUDY, OTHER_STUDY]
EVENING_OPTION = {CAR_STUDY: True, LAB_STUDY: False, OTHER_STUDY: True}
APP_OPTION = {CAR_STUDY: True, LAB_STUDY: False, OTHER_STUDY: True}

SEPARATOR = f"{'-' * 30}\n"
DEFAULT_BARCODE_FILE_SUFFIX = "_barcodes"
DEFAULT_QR_FILE_SUFFIX = "_qr_code"


@tui()
@click.command()
@click.option(
    "--study-name", required=True, prompt="Study name", type=str, help="A descriptive abbreviation for your study name."
)
@click.option(
    "--num-days", required=True, prompt="Duration of study [days]", type=int, help="The duration of your study in days."
)
@click.option(
    "--subject-data",
    required=True,
    prompt="Read subject IDs from file?",
    help="Whether the subject IDs should be read from a file.",
    is_flag=True,
)
@click.option(
    "--subject-path",
    default=None,
    required=False,
    prompt="Path to subject data file",
    help="The path to the *.csv or *.txt file with the participant data.",
    envvar="PATHS",
    type=click.Path(exists=True, file_okay=True, dir_okay=False),
    callback=validate_subject_path,
    cls=Condition,
    pos_condition="subject_data",
)
@click.option(
    "--subject-column",
    default="subject",
    required=False,
    prompt="Name of subject ID column",
    type=str,
    help="The name of the subject ID column in participants data file.",
    cls=Condition,
    pos_condition="subject_data",
)
@click.option(
    "--num-subjects",
    required=False,
    prompt="Number of participants",
    type=int,
    help="The number of participants in your study.",
    cls=Condition,
    neg_condition="subject_data",
)
@click.option(
    "--has-subject-prefix",
    prompt="Add prefix to participant number (e.g., 'VP_')?",
    help="Whether a prefix should be added to the participant number.",
    is_flag=True,
    cls=Condition,
    neg_condition="subject_path",
)
@click.option(
    "--subject-prefix",
    default="VP_",
    required=False,
    prompt="Subject prefix",
    type=str,
    help="Specification of subject prefix.",
    cls=Condition,
    pos_condition="has_subject_prefix",
)
@click.option(
    "--num-samples",
    required=True,
    prompt="Number of biomarker samples [during the day, WITHOUT evening sample]",
    type=int,
    help="The daily number of biomarker samples taken throughout the day from every participant. "
    "This number does not include the evening sample.",
)
@click.option(
    "--sample_prefix",
    required=True,
    default="S",
    prompt="Prefix for your type of biomarker (e.g., S for saliva)",
    type=str,
    help="The prefix for the biomarker type of the collected samples.",
)
@click.option(
    "--sample_start_id",
    required=True,
    default=0,
    prompt="Should biomarker IDs start at 0 or at 1 (i.e., S0 vs. S1)?",
    help="Whether the biomarker IDs should start at 0 or at 1.",
    type=click.IntRange(0, 1),
)
@click.option(
    "--study_type",
    required=True,
    default=CAR_STUDY,
    prompt="Type of study? (1 for CAR study, 2 for Lab-based study, 3 for Other)?",
    help="The type of study. 1 for CAR study, 2 for Lab-based study, 3 for Other.",
    type=click.IntRange(1, len(STUDY_TYPES)),
)
@click.option(
    "--has-evening-sample",
    prompt="Additional evening sample taken?",
    is_flag=True,
    help="Whether an additional biomarker sample is taken in the evening.",
    cls=NumericChoice,
    chosen_number="study_type",
    option_map=EVENING_OPTION,
)
@click.option(
    "--generate-barcode",
    required=True,
    prompt=f"{SEPARATOR}Generate printable labels for study?",
    is_flag=True,
    help="Whether a PDF with barcodes encoding the information for individual biomarker samples should be generated.",
)
@click.option(
    "--add-name",
    prompt="Add study name and participant number to label?",
    is_flag=True,
    help="Whether a the study name and participant id will be printed on every individual label.",
    cls=Condition,
    pos_condition="generate_barcode",
)
@click.option(
    "--has-barcode",
    prompt="Add barcode to label?",
    is_flag=True,
    help="Whether a barcode encoding the participant id, day of study, and number of biomarker sample will be"
    " printed on every individual label.",
    cls=Condition,
    pos_condition="generate_barcode",
)
@click.option(
    "--default-layout",
    prompt="Use Avery Zweckform J4791 as default layout?",
    is_flag=True,
    default=True,
    help="Whether the default layout will be used specified.",
    cls=Condition,
    pos_condition="generate_barcode",
)
@click.option(
    "--num_cols",
    prompt="Number of columns",
    type=int,
    help="The number of distinct labels per column",
    cls=Condition,
    neg_condition="default_layout",
)
@click.option(
    "--num_rows",
    prompt="Number of rows",
    type=int,
    help="The number of distinct labels per row",
    cls=Condition,
    neg_condition="default_layout",
)
@click.option(
    "--left_margin",
    prompt="Left margin",
    type=float,
    help="The offset between edge of sheet and first label to the left in mm",
    cls=Condition,
    neg_condition="default_layout",
)
@click.option(
    "--right_margin",
    prompt="Right margin",
    type=float,
    help="The offset between edge of sheet and first label to the right in mm",
    cls=Condition,
    neg_condition="default_layout",
)
@click.option(
    "--top_margin",
    prompt="Top margin",
    type=float,
    help="The offset between edge of sheet and first label to the top in mm",
    cls=Condition,
    neg_condition="default_layout",
)
@click.option(
    "--bottom_margin",
    prompt="Bottom margin",
    type=float,
    help="The offset between edge of sheet and first label to the bottom in mm",
    cls=Condition,
    neg_condition="default_layout",
)
@click.option(
    "--inter_col",
    prompt="Distance between columns",
    type=float,
    help="The distance between each label along the columns in mm",
    cls=Condition,
    neg_condition="default_layout",
)
@click.option(
    "--inter_row",
    prompt="Distance between rows",
    type=float,
    help="The distance between each label along the rows in mm",
    cls=Condition,
    neg_condition="default_layout",
)
@click.option(
    "--output_name_label",
    default=lambda: get_file_name(DEFAULT_BARCODE_FILE_SUFFIX),
    prompt="Name of label file",
    help="Name of the generated label file.",
    envvar="PATHS",
    type=click.Path(file_okay=True, dir_okay=False),
    cls=Condition,
    pos_condition="generate_barcode",
)
@click.option(
    "--generate-qr",
    prompt=f"{SEPARATOR}Use CAR Watch app for study?",
    is_flag=True,
    cls=NumericChoice,
    chosen_number="study_type",
    option_map=APP_OPTION,
    help="Whether a qr code encoding the study data for configuring the CAR watch app should be generated.",
)
@click.option(
    "--output_name_qr",
    default=lambda: get_file_name(DEFAULT_QR_FILE_SUFFIX),
    prompt="Name of QR code file",
    help="Name of the generated QR code file.",
    envvar="PATHS",
    type=click.Path(file_okay=True, dir_okay=False),
    cls=Condition,
    pos_condition="generate_qr",
)
@click.option(
    "--saliva-distances",
    default="15",
    required=False,
    prompt="Please specify duration between all saliva samples in minutes"
    " (as number when constant, as comma-separated list when varying from sample to sample)",
    type=str,
    help="The duration between saliva samples in minutes.",
    callback=validate_saliva_distances,
    cls=Condition,
    pos_condition="generate_qr",
)
@click.option(
    "--contact-email",
    required=False,
    prompt="Contact E-Mail Address that should receive CARWatch app timestamps",
    type=str,
    help="The E-Mail Address that will be used as default when sharing data from CARWatch app.",
    callback=validate_mail_input,
    cls=Condition,
    pos_condition="generate_qr",
)
@click.option(
    "--check-duplicates",
    prompt="Check for duplicate barcodes in app?",
    default=False,
    is_flag=True,
    cls=Condition,
    pos_condition="generate_qr",
    help="Whether the CARWatch app will check for every barcode, if it was scanned before.",
)
@click.option(
    "--enable-manual-scan",
    prompt="Enable manual scanning in app?",
    default=False,
    is_flag=True,
    cls=Condition,
    pos_condition="generate_qr",
    help="Whether the CARWatch app will allow manual scanning of sample barcodes apart from timed alarms.",
)
@click.option(
    "--output_dir",
    default=".",
    prompt="Output directory for generated files",
    help="Directory where generated files will be stored.",
    envvar="PATHS",
    type=click.Path(exists=True, file_okay=False, dir_okay=True),
)
def run(
    sample_prefix: Optional[str] = None,
    study_name: Optional[str] = None,
    num_days: Optional[int] = None,
    num_samples: Optional[int] = None,
    subject_path: Optional[str] = None,
    subject_column: Optional[str] = None,
    num_subjects: Optional[int] = None,
    subject_prefix: Optional[str] = None,
    has_evening_sample: Optional[bool] = None,
    **kwargs,
):
    """Generate barcode labels and QR codes for CAR study.

    Parameters
    ----------
    sample_prefix : str, optional
        Prefix for sample barcodes. Default: None
    study_name : str, optional
        Name of the study. Default: None
    num_days : int, optional
        Number of days of the study. Default: None
    num_samples : int, optional
        Number of samples per day. Default: None
    subject_path : str, optional
        Path to the subject file. Default: None
    subject_column : str, optional
        Name of the column in the subject file that contains the subject IDs. Default: None
    num_subjects : int, optional
        Number of subjects. Default: None
    subject_prefix : str, optional
        Prefix for subject IDs. Default: None
    has_evening_sample : bool, optional
        Whether the study has an evening sample. Default: None
    **kwargs : dict
        Additional keyword arguments.

    """
    done = False

    def animate():
        """Create a loading icon while input is processed."""
        for c in itertools.cycle(["|", "/", "-", "\\"]):
            if done:
                break
            sys.stdout.write("\rloading " + c)
            sys.stdout.flush()
            time.sleep(0.1)

    t = threading.Thread(target=animate)
    t.start()

    if not kwargs["generate_qr"] and not kwargs["generate_barcode"]:
        done = True
        raise click.UsageError("Nothing to do, no output generated.")

    start_sample_from_zero = kwargs["sample_start_id"] == 0

    if not kwargs["has_subject_prefix"]:
        subject_prefix = ""

    study = Study(
        study_name=study_name,
        num_days=num_days,
        num_samples=num_samples,
        start_sample_from_zero=start_sample_from_zero,
        num_subjects=num_subjects,
        subject_path=subject_path,
        subject_column=subject_column,
        subject_prefix=subject_prefix,
        has_evening_sample=has_evening_sample,
        sample_prefix=sample_prefix,
    )

    if kwargs["generate_barcode"]:
        _generate_barcode(study, **kwargs)
    if kwargs["generate_qr"]:
        try:
            _generate_qr_code(study, **kwargs)
        except ValueError as e:
            done = True
            raise click.BadParameter(str(e)) from e

    done = True


def _generate_barcode(study: Study, **kwargs):
    generator = LabelGenerator(
        study=study,
        add_name=kwargs["add_name"],
        has_barcode=kwargs["has_barcode"],
    )
    if not kwargs["default_layout"]:
        layout = CustomLayout(
            num_cols=kwargs["num_cols"],
            num_rows=kwargs["num_rows"],
            left_margin=kwargs["left_margin"],
            right_margin=kwargs["right_margin"],
            top_margin=kwargs["top_margin"],
            bottom_margin=kwargs["bottom_margin"],
            inter_col=kwargs["inter_col"],
            inter_row=kwargs["inter_row"],
        )
        generator.generate(output_dir=kwargs["output_dir"], output_name=kwargs["output_name_label"], layout=layout)
    else:
        generator.generate(output_dir=kwargs["output_dir"], output_name=kwargs["output_name_label"])


def _generate_qr_code(study, **kwargs):
    saliva_distances = _parse_saliva_distances(kwargs["saliva_distances"])
    generator = QrCodeGenerator(
        study=study,
        saliva_distances=saliva_distances,
        contact_email=kwargs["contact_email"],
        check_duplicates=kwargs["check_duplicates"],
        enable_manual_scan=kwargs["enable_manual_scan"],
    )
    generator.generate(output_dir=kwargs["output_dir"], output_name=kwargs["output_name_qr"])


def _parse_saliva_distances(saliva_distances: Union[Sequence[str], str]) -> Union[int, Sequence[int]]:
    saliva_distances = saliva_distances.replace(" ", "")  # trim spaces
    # list of int
    if "," in saliva_distances:
        saliva_distances = [int(dist) for dist in saliva_distances.split(",")]  # pylint: disable=eval-used
        return saliva_distances

    return int(saliva_distances)


if __name__ == "__main__":
    run()
